
use axum::{
	body::{Bytes,Full},
	headers::HeaderValue,
	http::StatusCode,
	http::header,
	http::header::SET_COOKIE,
	response::IntoResponse,
	response::Response,
};
use tera::Tera;
use toml::Table;

use std::marker::PhantomData;

use crate::mycelium::MycView;
use crate::mycelium::MycFormat;
use crate::mycelium::MycFormatFamily;
use crate::mycelium::MycQuerySettings;

/* The engine itself */

#[derive(Clone)]
pub struct MycEngine<V, S, F>
where V: MycView<S, F>, S: MycQuerySettings<F>, F: MycFormat {
	pub tera: Tera,
	pub template_config: Option<Table>,

	phantom_view: PhantomData<V>,
	phantom_settings: PhantomData<S>,
	phantom_format: PhantomData<F>,
}

impl<V, S, F> MycEngine<V, S, F>
where V: MycView<S, F>, S: MycQuerySettings<F>, F: MycFormat {

	pub fn new(tera: Tera, template_config: Option<Table>) -> Self {
		Self {
			tera: tera,
			template_config: template_config,
			phantom_view: PhantomData,
			phantom_settings: PhantomData,
			phantom_format: PhantomData,
		}
	}

	/// Takes settings and a view, converting it to a serveable response.
	pub async fn render_view(
		&self,
		settings: &S,
		view: V,
	) -> Response {
		let status_code = view.get_status_code(&settings);
		let cookie_string = view.get_cookie_header(&settings);
		let format = settings.get_format();

		let mut response = match format.get_family() {
			MycFormatFamily::Template => {
				let template_name = view.get_template_name();
				let mime_type = format.get_mime_type();
		
				let mut context = tera::Context::new();
				context.insert("view", &template_name);
				//intented for shared macros
				context.insert("format", &format.get_name());
				context.insert("mimetype", &mime_type.to_string());
				context.insert("data", &view);
				context.insert("extra", &self.template_config);
				settings.initalize_template_context(&mut context);

				match self.tera.render(&(template_name.clone()+&format.get_file_extension()), &context) {
					Ok(text) => {
						let mut response = (
							[(
								header::CONTENT_TYPE,
								HeaderValue::from_str(mime_type.as_ref())
									.expect("MimeType should always be a valid header value.")
							)],
							Into::<Full<Bytes>>::into(text),
						).into_response();
						view.update_response(&mut response, &settings);
						response
					},
					Err(e) => {
						println!("There was an error while rendering template {}: {e:?}", template_name);
						(
							StatusCode::INTERNAL_SERVER_ERROR,
						format!("Template error in {}, contact owner or see logs.\n", template_name)
						).into_response()
					}
				}
			}
			MycFormatFamily::API => {
				view.get_api_response(&settings)
			}
		};

		// Everything went well and nobody did the following work for us.
		if response.status() == StatusCode::OK {
			
			// Set status code
			if status_code != StatusCode::OK {
				*response.status_mut() = status_code;
			}

			// Set cookie header
			if let Some(cookie_string) = cookie_string {
				if let Ok(header_value) = HeaderValue::from_str(&cookie_string) {
					response.headers_mut().append(
						SET_COOKIE,
						header_value,
					);
				}
			}
			
		}

		// return response
		response
	}
}

