
use serde::Deserialize;
use serde::Serialize;
use mime::Mime;

/// Defines how the response should be rendered.
pub enum MycFormatFamily {
	/// When rendering the templating engine will be invoked
	Template,
	/// When rendering the [View](./trait.View.html)
	/// is asked to generate an API response.
	API,
}

/// Implement on a type that is able to describe a response format.
/// 
/// It is best implemented on an enum.
pub trait MycFormat: ToString+Clone+Default+Send {

	// Return the format family this 
	fn get_family(&self) -> MycFormatFamily {
		match self.get_name().as_str() {
			"json" => MycFormatFamily::API,
			_ => MycFormatFamily::Template,
		}
	}

	/// Returns the file extnsion for the format.
	///
	/// Used for deriving the path for the template name.
	///
	/// Defaults to `.{self.get_name()}`
	/// with the exception of the name being `text`
	/// then it defaults to `.txt`.
	fn get_file_extension(&self) -> String {
		match self.get_name().as_str() {
			"text" => ".txt".to_string(),
			_ => ".".to_owned()+&self.get_name(),
		}
	}

	/// Returns the name of the format,
	/// by default taken from the ToString implementation.
	fn get_name(&self) -> String {
		self.to_string()
	}

	/// Allows adding extra mimetypes quickly for prototyping
	///
	/// Implementing get_mime_type() properly is recommended
	/// for production use. 
	fn get_less_well_known_mimetype(&self) -> Option<Mime> {
		None
	}

	/// Returns a textual representation of the Mimetype.
	///
	/// It is recommended to implement this when in production use.
	///
	/// For prototyping the default implementation makes assumptions
	/// based on the output of get_name(), falling back
	/// to get_less_well_known_mimetype() and the "application/octet-stream" type.
	///
	/// The default implementation knows the following associations:
	/// * `text`: `text/plain; charset=utf-8`
	/// * `html`: `text/html; charset=utf-8`
	/// * `json`: `application/json`
	/// * `xml`: `application/xml`
	/// * `rss`: `application/rss+xml`
	/// * `atom`: `application/atom+xml`
	///
	/// *Implementation Note:* It may be possible that two different views
	/// have the same MimeType (maybe two json representations for different consumers).
	/// 
	fn get_mime_type(&self) -> Mime {
		match self.get_name().as_str() {
			"text" => mime::TEXT_PLAIN_UTF_8,
			"html" => mime::TEXT_HTML_UTF_8,
			"json" => mime::APPLICATION_JSON,
			"xml" => "application/xml".parse()
				.expect("Parse static application/xml"),
			"rss" => "application/rss+xml".parse()
				.expect("Parse static application/rss+xml"),
			"atom" =>  "application/atom+xml".parse()
				.expect("Parse static application/atom+xml"),
			_ =>
				self.get_less_well_known_mimetype()
					.unwrap_or(mime::APPLICATION_OCTET_STREAM),
		}
	}

	/// Constructs a view from its name.
	fn from_name(name: &str) -> Option<Self>;
}

// Some Sample implementations

#[derive(Clone,Serialize,Deserialize,Default)]
#[serde(rename_all="lowercase")]
pub enum HtmlTextJsonFormat {
	#[default]
	Html,
	Text,
	Json,
}

impl ToString for HtmlTextJsonFormat {
	fn to_string(&self) -> String {
		match self {
			Self::Html => "html",
			Self::Text => "text",
			Self::Json => "json",
		}.to_owned()
	}
}


impl MycFormat for HtmlTextJsonFormat {

	//TODO: implement other methods to make it more performant
	
	fn from_name(name: &str) -> Option<Self> {
		match name {
			"html" => Some(Self::Html),
			"text" => Some(Self::Text),
			"json" => Some(Self::Json),
			_ => None,
		}
	}
}


